@@ -10,13 +10,90 @@
 #include "eigenpy/fwd.hpp"
 
 #include <patchlevel.h> // For PY_MAJOR_VERSION
-#include <numpy/arrayobject.h>
 #include <iostream>
 
 #include "eigenpy/eigenpy.hpp"
 #include "eigenpy/registration.hpp"
 #include "eigenpy/map.hpp"
 
+namespace boost { namespace python { namespace detail {
+
+  template<class MatType>
+  struct referent_size<Eigen::MatrixBase<MatType>&>
+  {
+      BOOST_STATIC_CONSTANT(
+          std::size_t, value = sizeof(MatType));
+  };
+
+}}}
+
+namespace boost { namespace python { namespace converter {
+
+template<class MatType>
+struct implicit<Eigen::MatrixBase<MatType>,MatType>
+{
+  typedef Eigen::MatrixBase<MatType> Source;
+  typedef MatType Target;
+  
+  static void* convertible(PyObject* obj)
+  {
+    // Find a converter which can produce a Source instance from
+    // obj. The user has told us that Source can be converted to
+    // Target, and instantiating construct() below, ensures that
+    // at compile-time.
+    return implicit_rvalue_convertible_from_python(obj, registered<Source>::converters)
+    ? obj : 0;
+  }
+  
+  static void construct(PyObject* obj, rvalue_from_python_stage1_data* data)
+  {
+    void* storage = ((rvalue_from_python_storage<Target>*)data)->storage.bytes;
+    
+    arg_from_python<Source> get_source(obj);
+    bool convertible = get_source.convertible();
+    BOOST_VERIFY(convertible);
+    
+    new (storage) Target(get_source().derived());
+    
+    // record successful construction
+    data->convertible = storage;
+  }
+};
+
+template<class MatType>
+struct implicit<MatType,Eigen::MatrixBase<MatType> >
+{
+  typedef MatType Source;
+  typedef Eigen::MatrixBase<MatType> Target;
+  
+  static void* convertible(PyObject* obj)
+  {
+    // Find a converter which can produce a Source instance from
+    // obj. The user has told us that Source can be converted to
+    // Target, and instantiating construct() below, ensures that
+    // at compile-time.
+    return implicit_rvalue_convertible_from_python(obj, registered<Source>::converters)
+    ? obj : 0;
+  }
+  
+  static void construct(PyObject* obj, rvalue_from_python_stage1_data* data)
+  {
+    void* storage = reinterpret_cast<rvalue_from_python_storage<Target>*>
+                   (reinterpret_cast<void*>(data))->storage.bytes;
+    
+    arg_from_python<Source> get_source(obj);
+    bool convertible = get_source.convertible();
+    BOOST_VERIFY(convertible);
+    
+    new (storage) Source(get_source());
+    
+    // record successful construction
+    data->convertible = storage;
+  }
+};
+
+}}} // namespace boost::python::converter
+
 #define GET_PY_ARRAY_TYPE(array) PyArray_ObjectType(reinterpret_cast<PyObject *>(array), 0)
 
 namespace eigenpy
@@ -119,6 +196,28 @@ namespace eigenpy
     {
       return getInstance().CurrentNumpyType;
     }
+    
+    static const PyTypeObject * getNumpyMatrixType()
+    {
+      return getInstance().NumpyMatrixType;
+    }
+    
+    static const PyTypeObject * getNumpyArrayType()
+    {
+      return getInstance().NumpyArrayType;
+    }
+    
+    static bool isMatrix()
+    {
+      return PyType_IsSubtype(reinterpret_cast<PyTypeObject*>(getInstance().CurrentNumpyType.ptr()),
+                              getInstance().NumpyMatrixType);
+    }
+    
+    static bool isArray()
+    {
+      return PyType_IsSubtype(reinterpret_cast<PyTypeObject*>(getInstance().CurrentNumpyType.ptr()),
+                              getInstance().NumpyArrayType);
+    }
 
   protected:
     NumpyType()
@@ -150,6 +249,39 @@ namespace eigenpy
     bp::object NumpyArrayObject; PyTypeObject * NumpyArrayType;
     
   };
+
+  template<typename MatType, bool IsVectorAtCompileTime = MatType::IsVectorAtCompileTime>
+  struct initEigenObject
+  {
+    static MatType * run(PyArrayObject * pyArray, void * storage)
+    {
+      assert(PyArray_NDIM(pyArray) == 2);
+
+      const int rows = (int)PyArray_DIMS(pyArray)[0];
+      const int cols = (int)PyArray_DIMS(pyArray)[1];
+      
+      return new (storage) MatType(rows,cols);
+    }
+  };
+
+  template<typename MatType>
+  struct initEigenObject<MatType,true>
+  {
+    static MatType * run(PyArrayObject * pyArray, void * storage)
+    {
+      if(PyArray_NDIM(pyArray) == 1)
+      {
+        const int rows_or_cols = (int)PyArray_DIMS(pyArray)[0];
+        return new (storage) MatType(rows_or_cols);
+      }
+      else
+      {
+        const int rows = (int)PyArray_DIMS(pyArray)[0];
+        const int cols = (int)PyArray_DIMS(pyArray)[1];
+        return new (storage) MatType(rows,cols);
+      }
+    }
+  };
   
   template<typename MatType>
   struct EigenObjectAllocator
@@ -159,10 +291,7 @@ namespace eigenpy
     
     static void allocate(PyArrayObject * pyArray, void * storage)
     {
-      const int rows = (int)PyArray_DIMS(pyArray)[0];
-      const int cols = (int)PyArray_DIMS(pyArray)[1];
-      
-      Type * mat_ptr = new (storage) Type(rows,cols);
+      Type * mat_ptr = initEigenObject<Type>::run(pyArray,storage);
       
       if(NumpyEquivalentType<Scalar>::type_code == GET_PY_ARRAY_TYPE(pyArray))
       {
@@ -294,124 +423,149 @@ namespace eigenpy
   template<typename MatType>
   struct EigenFromPy
   {
+    
+    static bool isScalarConvertible(const int np_type)
+    {
+      if(NumpyEquivalentType<typename MatType::Scalar>::type_code == np_type)
+        return true;
+      
+      switch(np_type)
+      {
+        case NPY_INT:
+          return FromTypeToType<int,typename MatType::Scalar>::value;
+        case NPY_LONG:
+          return FromTypeToType<long,typename MatType::Scalar>::value;
+        case NPY_FLOAT:
+          return FromTypeToType<float,typename MatType::Scalar>::value;
+        case NPY_DOUBLE:
+          return FromTypeToType<double,typename MatType::Scalar>::value;
+        default:
+          return false;
+      }
+    }
+    
     /// \brief Determine if pyObj can be converted into a MatType object
     static void* convertible(PyArrayObject* pyArray)
     {
       if(!PyArray_Check(pyArray))
         return 0;
+      
+      if(!isScalarConvertible(GET_PY_ARRAY_TYPE(pyArray)))
+        return 0;
 
       if(MatType::IsVectorAtCompileTime)
       {
-        // Special care of scalar matrix of dimension 1x1.
-        if(PyArray_DIMS(pyArray)[0] == 1 && PyArray_DIMS(pyArray)[1] == 1)
-          return pyArray;
+        const Eigen::DenseIndex size_at_compile_time
+        = MatType::IsRowMajor
+        ? MatType::ColsAtCompileTime
+        : MatType::RowsAtCompileTime;
         
-        if(PyArray_DIMS(pyArray)[0] > 1 && PyArray_DIMS(pyArray)[1] > 1)
+        switch(PyArray_NDIM(pyArray))
         {
+          case 0:
+            return 0;
+          case 1:
+          {
+            if(size_at_compile_time != Eigen::Dynamic)
+            {
+              // check that the sizes at compile time matche
+              if(PyArray_DIMS(pyArray)[0] == size_at_compile_time)
+                return pyArray;
+              else
+                return 0;
+            }
+            else // This is a dynamic MatType
+              return pyArray;
+          }
+          case 2:
+          {
+            // Special care of scalar matrix of dimension 1x1.
+            if(PyArray_DIMS(pyArray)[0] == 1 && PyArray_DIMS(pyArray)[1] == 1)
+            {
+              if(size_at_compile_time != Eigen::Dynamic)
+              {
+                if(size_at_compile_time == 1)
+                  return pyArray;
+                else
+                  return 0;
+              }
+              else // This is a dynamic MatType
+                return pyArray;
+            }
+            
+            if(PyArray_DIMS(pyArray)[0] > 1 && PyArray_DIMS(pyArray)[1] > 1)
+            {
 #ifndef NDEBUG
-          std::cerr << "The number of dimension of the object does not correspond to a vector" << std::endl;
+              std::cerr << "The number of dimension of the object does not correspond to a vector" << std::endl;
 #endif
-          return 0;
+              return 0;
+            }
+            
+            if(((PyArray_DIMS(pyArray)[0] == 1) && (MatType::ColsAtCompileTime == 1))
+               || ((PyArray_DIMS(pyArray)[1] == 1) && (MatType::RowsAtCompileTime == 1)))
+            {
+#ifndef NDEBUG
+              if(MatType::ColsAtCompileTime == 1)
+                std::cerr << "The object is not a column vector" << std::endl;
+              else
+                std::cerr << "The object is not a row vector" << std::endl;
+#endif
+              return 0;
+            }
+            
+            if(size_at_compile_time != Eigen::Dynamic)
+            { // This is a fixe size vector
+              const Eigen::DenseIndex pyArray_size
+              = PyArray_DIMS(pyArray)[0] > PyArray_DIMS(pyArray)[1]
+              ? PyArray_DIMS(pyArray)[0]
+              : PyArray_DIMS(pyArray)[1];
+              if(size_at_compile_time != pyArray_size)
+                return 0;
+            }
+            break;
+          }
+          default:
+            return 0;
+        }
+      }
+      else // this is a matrix
+      {
+        if(PyArray_NDIM(pyArray) != 2)
+        {
+          if ( (PyArray_NDIM(pyArray) !=1) || (! MatType::IsVectorAtCompileTime) )
+          {
+#ifndef NDEBUG
+            std::cerr << "The number of dimension of the object is not correct." << std::endl;
+#endif
+            return 0;
+          }
         }
         
-        if(((PyArray_DIMS(pyArray)[0] == 1) && (MatType::ColsAtCompileTime == 1))
-           || ((PyArray_DIMS(pyArray)[1] == 1) && (MatType::RowsAtCompileTime == 1)))
+        if(PyArray_NDIM(pyArray) == 2)
         {
-#ifndef NDEBUG
-          if(MatType::ColsAtCompileTime == 1)
-            std::cerr << "The object is not a column vector" << std::endl;
-          else
-            std::cerr << "The object is not a row vector" << std::endl;
-#endif
-          return 0;
+          const int R = (int)PyArray_DIMS(pyArray)[0];
+          const int C = (int)PyArray_DIMS(pyArray)[1];
+          
+          if( (MatType::RowsAtCompileTime!=R)
+             && (MatType::RowsAtCompileTime!=Eigen::Dynamic) )
+            return 0;
+          if( (MatType::ColsAtCompileTime!=C)
+             && (MatType::ColsAtCompileTime!=Eigen::Dynamic) )
+            return 0;
         }
       }
-      
-      if(PyArray_NDIM(pyArray) != 2)
-      {
-        if ( (PyArray_NDIM(pyArray) !=1) || (! MatType::IsVectorAtCompileTime) )
-        {
-#ifndef NDEBUG
-          std::cerr << "The number of dimension of the object is not correct." << std::endl;
-#endif
-          return 0;
-        }
-      }
-      
-      if(PyArray_NDIM(pyArray) == 2)
-      {
-        const int R = (int)PyArray_DIMS(pyArray)[0];
-        const int C = (int)PyArray_DIMS(pyArray)[1];
         
-        if( (MatType::RowsAtCompileTime!=R)
-           && (MatType::RowsAtCompileTime!=Eigen::Dynamic) )
-          return 0;
-        if( (MatType::ColsAtCompileTime!=C)
-           && (MatType::ColsAtCompileTime!=Eigen::Dynamic) )
-          return 0;
-      }
-      
-      // Check if the Scalar type of the obj_ptr is compatible with the Scalar type of MatType
-      if(GET_PY_ARRAY_TYPE(pyArray) == NPY_INT)
-      {
-        if(!FromTypeToType<int,typename MatType::Scalar>::value)
-        {
-#ifndef NDEBUG
-          std::cerr << "The Python matrix scalar type (int) cannot be converted into the scalar type of the Eigen matrix. Loss of arithmetic precision" << std::endl;
-#endif
-          return 0;
-        }
-      }
-      else if(GET_PY_ARRAY_TYPE(pyArray) == NPY_LONG)
-      {
-        if(!FromTypeToType<long,typename MatType::Scalar>::value)
-        {
-#ifndef NDEBUG
-          std::cerr << "The Python matrix scalar type (long) cannot be converted into the scalar type of the Eigen matrix. Loss of arithmetic precision" << std::endl;
-#endif
-          return 0;
-        }
-      }
-      else if(GET_PY_ARRAY_TYPE(pyArray) == NPY_FLOAT)
-      {
-        if(!FromTypeToType<float,typename MatType::Scalar>::value)
-        {
-#ifndef NDEBUG
-          std::cerr << "The Python matrix scalar type (float) cannot be converted into the scalar type of the Eigen matrix. Loss of arithmetic precision" << std::endl;
-#endif
-          return 0;
-        }
-      }
-      else if(GET_PY_ARRAY_TYPE(pyArray) == NPY_DOUBLE)
-      {
-        if(!FromTypeToType<double,typename MatType::Scalar>::value)
-        {
-#ifndef NDEBUG
-          std::cerr << "The Python matrix scalar (double) type cannot be converted into the scalar type of the Eigen matrix. Loss of arithmetic precision." << std::endl;
-#endif
-          return 0;
-        }
-      }
-      else if(GET_PY_ARRAY_TYPE(pyArray) != NumpyEquivalentType<typename MatType::Scalar>::type_code)
-      {
-#ifndef NDEBUG
-        std::cerr << "The internal type as no Eigen equivalent." << std::endl;
-#endif
-        
-        return 0;
-      }
-      
 #ifdef NPY_1_8_API_VERSION
       if(!(PyArray_FLAGS(pyArray)))
 #else
       if(!(PyArray_FLAGS(pyArray) & NPY_ALIGNED))
 #endif
-        {
+      {
 #ifndef NDEBUG
-          std::cerr << "NPY non-aligned matrices are not implemented." << std::endl;
+        std::cerr << "NPY non-aligned matrices are not implemented." << std::endl;
 #endif
-          return 0;
-        }
+        return 0;
+      }
       
       return pyArray;
     }
@@ -423,8 +577,8 @@ namespace eigenpy
       PyArrayObject * pyArray = reinterpret_cast<PyArrayObject*>(pyObj);
       assert((PyArray_DIMS(pyArray)[0]<INT_MAX) && (PyArray_DIMS(pyArray)[1]<INT_MAX));
       
-      void* storage = ((bp::converter::rvalue_from_python_storage<MatType>*)
-                       ((void*)memory))->storage.bytes;
+      void* storage = reinterpret_cast<bp::converter::rvalue_from_python_storage<MatType>*>
+                     (reinterpret_cast<void*>(memory))->storage.bytes;
       
       EigenObjectAllocator<MatType>::allocate(pyArray,storage);
 
@@ -439,36 +593,8 @@ namespace eigenpy
     }
   };
   
-  template<typename MatType>
-  struct EigenFromPy< Eigen::MatrixBase<MatType> >
-  {
-    typedef EigenFromPy<MatType> EigenFromPyDerived;
-    typedef Eigen::MatrixBase<MatType> Base;
-    
-    /// \brief Determine if pyObj can be converted into a MatType object
-    static void* convertible(PyArrayObject* pyObj)
-    {
-      return EigenFromPyDerived::convertible(pyObj);
-    }
-    
-    /// \brief Allocate memory and copy pyObj in the new storage
-    static void construct(PyObject* pyObj,
-                          bp::converter::rvalue_from_python_stage1_data* memory)
-    {
-      EigenFromPyDerived::construct(pyObj,memory);
-    }
-    
-    static void registration()
-    {
-      bp::converter::registry::push_back
-      (reinterpret_cast<void *(*)(_object *)>(&EigenFromPy::convertible),
-       &EigenFromPy::construct,bp::type_id<Base>());
-    }
-  };
-  
-#define numpy_import_array() {if (_import_array() < 0) {PyErr_Print(); PyErr_SetString(PyExc_ImportError, "numpy.core.multiarray failed to import"); } }
-  
   template<typename MatType,typename EigenEquivalentType>
+  EIGENPY_DEPRECATED
   void enableEigenPySpecific()
   {
     enableEigenPySpecific<MatType>();
@@ -483,12 +609,13 @@ namespace eigenpy
 
       // Add also conversion to Eigen::MatrixBase<MatType>
       typedef Eigen::MatrixBase<MatType> MatTypeBase;
-      EigenFromPy<MatTypeBase>::registration();
+//      bp::implicitly_convertible<MatTypeBase,MatType>();
+      bp::implicitly_convertible<MatType,MatTypeBase>();
     }
   };
 
 #if EIGEN_VERSION_AT_LEAST(3,2,0)
-  /// Template specialization for Eigen::Ref
+  // Template specialization for Eigen::Ref
   template<typename MatType>
   struct EigenFromPyConverter< eigenpy::Ref<MatType> >
   {
@@ -500,6 +627,8 @@ namespace eigenpy
     }
   };
 #endif
+
+#define numpy_import_array() {if (_import_array() < 0) {PyErr_Print(); PyErr_SetString(PyExc_ImportError, "numpy.core.multiarray failed to import"); } }
   
   template<typename MatType>
   void enableEigenPySpecific()
@@ -509,7 +638,6 @@ namespace eigenpy
     
     bp::to_python_converter<MatType,EigenToPy<MatType> >();
     EigenFromPyConverter<MatType>::registration();
-   
   }
 
 } // namespace eigenpy
